#include "version.h"

#ifdef USE_CLOUDSKY

#include "wl_def.h"
#include "wl_cloudsky.h"

// Each colormap defines a number of colors which should be mapped from
// the skytable. The according colormapentry_t array defines how these colors should
// be mapped to the wolfenstein palette. The first int of each entry defines
// how many colors are grouped to this entry and the absolute value of the
// second int sets the starting palette index for this pair. If this value is
// negative the index will be decremented for every color, if it's positive
// it will be incremented.
//
// Example colormap:
//   colormapentry_t colmapents_1[] = { { 6, -10 }, { 2, 40 } };
//   colormap_t colorMaps[] = {
//      { 8, colmapents_1 }
//   };
//
//   The colormap 0 consists of 8 colors. The first color group consists of 6
//   colors and starts descending at palette index 10: 10, 9, 8, 7, 6, 5
//   The second color group consists of 2 colors and starts ascending at
//   index 40: 40, 41
//   There's no other color group because all colors of this colormap are
//   already used (6+2=8)
//
// Warning: Always make sure that the sum of the amount of the colors in all
//          color groups is the number of colors used for your colormap!

colormapentry_t colmapents_1[] = { { 16, -31 }, { 16, 136 } };          //blue and white clouds - solid
colormapentry_t colmapents_2[] = { { 16, -31 } };                       //grey and white clouds - solid
colormapentry_t colmapents_3[] = { { 4, 0 }, { 3, 29 } };               //dark grey clouds      - see through
colormapentry_t colmapents_4[] = { { 20, -28 }, {4,-111} };             //white and green weird - see through
colormapentry_t colmapents_5[] = { { 4, 0 }, { 5, 217 } };              //dark brown red clouds - see through

colormap_t colorMaps[] = {
    { 32, colmapents_1 }, 
    { 16, colmapents_2 },
    { 7, colmapents_3 },
    { 24, colmapents_4 },
    { 9, colmapents_5 }
};

const int numColorMaps = lengthof(colorMaps);

// The sky definitions which can be selected as defined by GetCloudSkyDefID() in wl_def.h
// You can use <TAB>+Z in debug mode to find out suitable values for seed and colorMapIndex
// Each entry consists of seed, speed, angle and colorMapIndex
cloudsky_t cloudSkys[] = {
    { 626,   0,  20,  2 },    //01
    { 1234,  0,  60,  1 },    //02
    { 0,     0,  120, 2 },    //03
    { 0,     0,    0,   0 },    //04
    { 11243, 0,  310, 0 },    //05
    { 32141, 0,  87,  0 },    //06
    { 12124, 0,  64,  0 },    //07
    { 55543, 0,  240, 0 },    //08
    { 65535, 0,  54,  1 },    //09
    { 4,     990, 290, 4 },    //10    (31) Le Chateu    
    { 4,     0, 290, 0 },    //11
    { 626,   0,  20,  0 },    //12
    { 1234,  0,  60,  1 },    //13
    { 0,     0,  120, 0 },    //14
    { 0,     0,    0,   0 },    //15
    { 11243, 0,  310, 0 },    //16
    { 32141, 0,  87,  0 },    //17
    { 12124, 0,  64,  0 },    //18
    { 55543, 0,  240, 0 },    //19
    { 65535, 0,  54,  2 },    //20
    { 626,   660,  20,  2 },    //21    (21) DARK WOODS
    { 1234,  0,  60,  2 },    //22
    { 0,     0,  120, 0 },    //23
    { 0,     0,    0,   0 },    //24
    { 11243, 0,  310, 0 },    //25
    { 32141, 0,  87,  0 },    //26
    { 12124, 0,  64,  0 },    //27
    { 55543, 0,  240, 0 },    //28
    { 65535, 0,  54,  1 },    //29
    { 4,     0, 290, 0 },    //30
    { 8,     0, 290, 0 },    //31
    { 626,   0,  20,  0 },    //32
    { 1234,  0,  60,  1 },    //33
    { 0,     0,  120, 0 },    //34
    { 0,     0,    0,   0 },    //35
    { 11243, 0,  310, 0 },    //36
    { 32141, 0,  87,  0 },    //37
    { 12124, 0,  64,  0 },    //38
    { 55543, 0,  240, 0 },    //39
    { 65535, 0,  54,  3 },    //40
    { 626,   0,  20,  3 },    //41
    { 1234,  0,  60,  1 },    //42
    { 0,     0,  120, 0 },    //43
    { 0,     0,    0,   0 },    //44
    { 11243, 0,  310, 0 },    //45
    { 32141, 0,  87,  0 },    //46
    { 12124, 0,  64,  0 },    //47
    { 55543, 0,  240, 0 },    //48
    { 65535, 0,  54,  1 },    //49
    { 4,     0, 290, 0 },    //50
    { 4,     0, 290, 0 },    //51 ???
    { 626,   0,  20,  0 },    //52
    { 1234,  0,  60,  1 },    //53
    { 0,     0,  120, 0 },    //54
    { 0,     0,    0,   0 },    //55
    { 11243, 0,  310, 0 },    //56
    { 32141, 0,  87,  0 },    //57
    { 12124, 0,  64,  0 },    //58
    { 55543, 0,  240, 0 },    //59
    { 65535, 0,  54,  3 },    //60
  };

byte skyc[65536L];

long cloudx = 0, cloudy = 0;
cloudsky_t *curSky = NULL;

//#ifdef USE_FEATUREFLAGS

// The lower left tile of every map determines the used cloud sky definition from cloudSkys.
//static int GetCloudSkyDefID()
//{
//    int skyID = ffDataBottomLeft;
//    assert(skyID >= 0 && skyID < lengthof(cloudSkys));
//    return skyID;
//}

//#else

static int GetCloudSkyDefID()
{
    int skyID;
    switch(gamestate.episode * 10 + mapon)
    {
        case  0: skyID =  0; break;
        case  1: skyID =  1; break;
        case  2: skyID =  2; break;
        case  3: skyID =  3; break;
        case  4: skyID =  4; break;
        case  5: skyID =  5; break;
        case  6: skyID =  6; break;
        case  7: skyID =  7; break;
        case  8: skyID =  8; break;
        case  9: skyID =  9; break;
        case  20: skyID =  20; break;
        //case  21: skyID =  21; break;
        //case  22: skyID =  22; break;
        default: skyID =  9; break;
    }
    assert(skyID >= 0 && skyID < lengthof(cloudSkys));
    return skyID;
}


//#endif

void SplitS(unsigned size,unsigned x1,unsigned y1,unsigned x2,unsigned y2)
{
   if(size==1) return;
   if(!skyc[((x1+size/2)*256+y1)])
   {
      skyc[((x1+size/2)*256+y1)]=(byte)(((int)skyc[(x1*256+y1)]
            +(int)skyc[((x2&0xff)*256+y1)])/2)+rand()%(size*2)-size;
      if(!skyc[((x1+size/2)*256+y1)]) skyc[((x1+size/2)*256+y1)]=1;
   }
   if(!skyc[((x1+size/2)*256+(y2&0xff))])
   {
      skyc[((x1+size/2)*256+(y2&0xff))]=(byte)(((int)skyc[(x1*256+(y2&0xff))]
            +(int)skyc[((x2&0xff)*256+(y2&0xff))])/2)+rand()%(size*2)-size;
      if(!skyc[((x1+size/2)*256+(y2&0xff))])
         skyc[((x1+size/2)*256+(y2&0xff))]=1;
   }
   if(!skyc[(x1*256+y1+size/2)])
   {
      skyc[(x1*256+y1+size/2)]=(byte)(((int)skyc[(x1*256+y1)]
            +(int)skyc[(x1*256+(y2&0xff))])/2)+rand()%(size*2)-size;
      if(!skyc[(x1*256+y1+size/2)]) skyc[(x1*256+y1+size/2)]=1;
   }
   if(!skyc[((x2&0xff)*256+y1+size/2)])
   {
      skyc[((x2&0xff)*256+y1+size/2)]=(byte)(((int)skyc[((x2&0xff)*256+y1)]
            +(int)skyc[((x2&0xff)*256+(y2&0xff))])/2)+rand()%(size*2)-size;
      if(!skyc[((x2&0xff)*256+y1+size/2)]) skyc[((x2&0xff)*256+y1+size/2)]=1;
   }

   skyc[((x1+size/2)*256+y1+size/2)]=(byte)(((int)skyc[(x1*256+y1)]
         +(int)skyc[((x2&0xff)*256+y1)]+(int)skyc[(x1*256+(y2&0xff))]
         +(int)skyc[((x2&0xff)*256+(y2&0xff))])/4)+rand()%(size*2)-size;

   SplitS(size/2,x1,y1+size/2,x1+size/2,y2);
   SplitS(size/2,x1+size/2,y1,x2,y1+size/2);
   SplitS(size/2,x1+size/2,y1+size/2,x2,y2);
   SplitS(size/2,x1,y1,x1+size/2,y1+size/2);
}

void InitSky()
{
     
    unsigned cloudskyid = GetCloudSkyDefID();
    if(cloudskyid >= lengthof(cloudSkys))
        Quit("Illegal cloud sky id: %u", cloudskyid);
    curSky = &cloudSkys[cloudskyid];

    memset(skyc, 0, sizeof(skyc));
    // funny water texture if used instead of memset ;D
    // for(int i = 0; i < 65536; i++)
    //     skyc[i] = rand() % 32 * 8;

    srand(curSky->seed);
    skyc[0] = rand() % 256;
    SplitS(256, 0, 0, 256, 256);

    // Smooth the clouds a bit
    for(int k = 0; k < 2; k++)
    {
        for(int i = 0; i < 256; i++)
        {
            for(int j = 0; j < 256; j++)
            {
                int32_t val = -skyc[j * 256 + i];
                for(int m = 0; m < 3; m++)
                {
                    for(int n = 0; n < 3; n++)
                    {
                        val += skyc[((j + n - 1) & 0xff) * 256 + ((i + m - 1) & 0xff)];
                    }
                }
                skyc[j * 256 + i] = (byte)(val >> 3);
            }
        }
    }

    // the following commented line could be useful, if you're trying to
    // create a new color map. This will display your current color map
    // in one (of course repeating) stripe of the sky

     //for(int i = 0; i < 256; i++)
     //    skyc[i] = skyc[i + 256] = skyc[i + 512] = i;

    if(curSky->colorMapIndex >= lengthof(colorMaps))
        Quit("Illegal colorMapIndex for cloud sky def %u: %u", cloudskyid, curSky->colorMapIndex);

    colormap_t *curMap = &colorMaps[curSky->colorMapIndex];
    int numColors = curMap->numColors;
    byte colormap[256];
    colormapentry_t *curEntry = curMap->entries;
    for(int calcedCols = 0; calcedCols < numColors; curEntry++)
    {
        if(curEntry->startAndDir < 0)
        {
            for(int i = 0, ind = -curEntry->startAndDir; i < curEntry->length; i++, ind--)
                colormap[calcedCols++] = ind;
        }
        else
        {
            for(int i = 0, ind = curEntry->startAndDir; i < curEntry->length; i++, ind++)
                colormap[calcedCols++] = ind;
        }
    }

    for(int i = 0; i < 256; i++)
    {
        for(int j = 0; j < 256; j++)
        {
            skyc[i * 256 + j] = colormap[skyc[i * 256 + j] * numColors / 256];
        }
    }
}

// Based on Textured Floor and Ceiling by DarkOne
void DrawClouds(byte *vbuf, unsigned vbufPitch, int min_wallheight)
{
    // Move clouds
    fixed moveDist = tics * curSky->speed;
    cloudx += FixedMul(moveDist,sintable[curSky->angle]);
    cloudy -= FixedMul(moveDist,costable[curSky->angle]);

    // Draw them
    int y0, halfheight;
    unsigned top_offset0;
    fixed dist;                                // distance to row projection
    fixed tex_step;                            // global step per one screen pixel
    fixed gu, gv, du, dv;                      // global texture coordinates
    int u, v;                                  // local texture coordinates
    int skycindex;                             // added for see through clouds
  //  byte* curshades;

    // ------ * prepare * --------
    halfheight = viewheight >> 1;
    y0 = min_wallheight >> 3;                  // starting y value
    if(y0 > halfheight)
        return;                                // view obscured by walls
    if(!y0) y0 = 1;                            // don't let division by zero
    top_offset0 = vbufPitch * (halfheight - y0 - 1);

    // draw horizontal lines
    for(int y = y0, top_offset = top_offset0; y < halfheight; y++, top_offset -= vbufPitch)
    {
        dist = (heightnumerator / y) << 8;
        gu =  viewx + FixedMul(dist, viewcos) + cloudx;
        gv = -viewy + FixedMul(dist, viewsin) + cloudy;
        tex_step = (int32_t)((dist << 8) / viewwidth / (175 + totalExcursion * (4.25 / 1113.6))); 
        du =  FixedMul(tex_step, viewsin);
        dv = -FixedMul(tex_step, viewcos);
        gu -= (viewwidth >> 1)*du;
        gv -= (viewwidth >> 1)*dv;          // starting point (leftmost)
        for(int x = 0, top_add = top_offset; x < viewwidth; x++, top_add++)
        {
            if(wallheight[x] >> 3 <= y)
            {
                u = (gu >> 13) & 255;
                v = (gv >> 13) & 255;
                skycindex = ((255 - u) << 8) + 255 - v;
                //if (skyc[skycindex] <= 0x6f) // ff=pink or 255  //vbuf[top_add] = skyc[((255 - u) << 8) + 255 - v];
  //              if (skyc[skycindex] != (0xff||0x00||0x01||0x02||0x03||0x04||0x05||0x06||0x07||0x08||0x09||0x0a||0x0b||0x0c||0x0d||0x0e||0x0f))// || skyc[skycindex] >= 0xff) // ff=pink or 255  //vbuf[top_add] = skyc[((255 - u) << 8) + 255 - v];
                if (skyc[skycindex] >= 0x0f)
                {
                   #ifdef USE_SHADING
                          vbuf[top_add] = skyc[skycindex];//curshades[skyc[skycindex]];
                   #else
                          vbuf[top_add] = skyc[skycindex];
                   #endif
                }
            }
            gu += du;
            gv += dv;
        } 
    }
}

#endif
